#include <stdio.h>
#include <stdlib.h>

#include <coff.h>
#include <bw/boio.h>

size_t
coff_filehdr_fread(struct filehdr *filehdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!filehdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_be16(&filehdr[tot].f_magic,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&filehdr[tot].f_nscns,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&filehdr[tot].f_timdat, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&filehdr[tot].f_symptr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&filehdr[tot].f_nsyms,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&filehdr[tot].f_opthdr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&filehdr[tot].f_flags,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_aouthdr_fread(struct aouthdr *aouthdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!aouthdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_be16(&aouthdr[tot].magic,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&aouthdr[tot].vstamp,     1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].tsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].dsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].bsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].entry,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].text_start, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&aouthdr[tot].data_start, 1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_scnhdr_fread(struct scnhdr *scnhdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!scnhdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_8(scnhdr[tot].s_name, S_SCNNMLEN, f);
        scnhdr[tot].s_name[S_SCNNMLEN] = '\0';
        if (S_SCNNMLEN != ret)
            goto out;

        ret = bw_fread_be32(&scnhdr[tot].s_paddr,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_vaddr,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_size,    1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_scnptr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_relptr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_lnnoptr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&scnhdr[tot].s_nreloc,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&scnhdr[tot].s_nlnno,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&scnhdr[tot].s_flags,   1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_reloc_fread(struct reloc *reloc, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!reloc || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_be32(&reloc[tot].r_vaddr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&reloc[tot].r_symndx, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&reloc[tot].r_offset, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&reloc[tot].r_type,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&reloc[tot].r_stuff,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_lineno_fread(struct lineno *lineno, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!lineno || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_be32(&lineno[tot].l_paddr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be32(&lineno[tot].l_lnno,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_syment_fread(struct syment *syment, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!syment || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fread_8(syment[tot].n_name, N_SYMNMLEN, f);
        syment[tot].n_name[N_SYMNMLEN] = '\0';
        if (N_SYMNMLEN != ret)
            goto out;

        if (!syment[tot].n_zeroes) {
            syment[tot].n_offset = bw_be32_to_cpu(syment[tot].n_offset);
        }

        ret = bw_fread_be32(&syment[tot].n_value,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&syment[tot].n_scnum,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_be16(&syment[tot].n_type,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_8(   &syment[tot].n_sclass, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fread_8(   &syment[tot].n_numaux, 1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_filehdr_fwrite(const struct filehdr *filehdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!filehdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fwrite_be16(&filehdr[tot].f_magic,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&filehdr[tot].f_nscns,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&filehdr[tot].f_timdat, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&filehdr[tot].f_symptr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&filehdr[tot].f_nsyms,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&filehdr[tot].f_opthdr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&filehdr[tot].f_flags,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_aouthdr_fwrite(const struct aouthdr *aouthdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!aouthdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fwrite_be16(&aouthdr[tot].magic,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&aouthdr[tot].vstamp,     1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].tsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].dsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].bsize,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].entry,      1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].text_start, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&aouthdr[tot].data_start, 1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_scnhdr_fwrite(const struct scnhdr *scnhdr, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!scnhdr || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fwrite_8(scnhdr[tot].s_name, S_SCNNMLEN, f);
        if (S_SCNNMLEN != ret)
            goto out;

        ret = bw_fwrite_be32(&scnhdr[tot].s_paddr,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_vaddr,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_size,    1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_scnptr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_relptr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_lnnoptr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&scnhdr[tot].s_nreloc,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&scnhdr[tot].s_nlnno,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&scnhdr[tot].s_flags,   1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_reloc_fwrite(const struct reloc *reloc, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!reloc || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fwrite_be32(&reloc[tot].r_vaddr,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&reloc[tot].r_symndx, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&reloc[tot].r_offset, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&reloc[tot].r_type,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&reloc[tot].r_stuff,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_lineno_fwrite(const struct lineno *lineno, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!lineno || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        ret = bw_fwrite_be32(&lineno[tot].l_paddr, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be32(&lineno[tot].l_lnno,  1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

size_t
coff_syment_fwrite(const struct syment *syment, size_t nmemb, FILE *f)
{
    size_t ret, tot = 0;

    if (!syment || !nmemb || !f)
        goto out;

    for (tot = 0; tot < nmemb; tot++) {
        if (syment[tot].n_zeroes) {
            ret = bw_fwrite_8(syment[tot].n_name, N_SYMNMLEN, f);
            if (N_SYMNMLEN != ret)
                goto out;
        } else {
            ret = bw_fwrite_be32(&syment[tot].n_zeroes, 1, f);
            if (1 != ret)
                goto out;
            ret = bw_fwrite_be32(&syment[tot].n_offset, 1, f);
            if (1 != ret)
                goto out;
        }


        ret = bw_fwrite_be32(&syment[tot].n_value,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&syment[tot].n_scnum,  1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_be16(&syment[tot].n_type,   1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_8(   &syment[tot].n_sclass, 1, f);
        if (1 != ret)
            goto out;
        ret = bw_fwrite_8(   &syment[tot].n_numaux, 1, f);
        if (1 != ret)
            goto out;
    }

out:
    return tot;
}

