/* -*- c -*- */

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>

#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <getopt.h>

#include <coff.h>

#include "utils.h"

int
main(int ac, char *av[])
{
    struct coff_file *coff = NULL;
    int i;
    int print_flags = COFF_FILEHDR | COFF_AOUTHDR;

    int opt_index = 0, ch;
    const char *optstring = "+asrldShc";
    static const struct option long_opt[] =
    {
        { "all",          0, 0, 'a' },
        { "sections",     0, 0, 's' },
        { "relocations",  0, 0, 'r' },
        { "line-numbers", 0, 0, 'l' },
        { "data",         0, 0, 'd' },
        { "symbol-table", 0, 0, 'S' },
        { "clear",        0, 0, 'c' },
        { 0, 0, 0, 0 }
    };

    while (
        (ch = getopt_long(ac, av, optstring, long_opt, &opt_index)) != -1
    ) {
        switch (ch) {
        case 'a':
            print_flags = ~0;
            break;
        case 's':
            print_flags |= COFF_SCNS;
            break;
        case 'r':
            print_flags |= COFF_SCNS | COFF_RELOC;
            break;
        case 'l':
            print_flags |= COFF_SCNS | COFF_LNNO;
            break;
        case 'd':
            print_flags |= COFF_SCNS | COFF_SCNSDATA;
            break;
        case 'S':
            print_flags |= COFF_SYMTAB;
            break;
        case 'c':
            print_flags = 0;
            break;
        default:
            exit(1);
        }
    }
    ac -= optind;
    av += optind;

    for (i = 0; i < ac; i++) {
        printf("%s:\n", av[i]);

        coff = read_coff(av[i]);

        print_coff(coff, print_flags);

        free_coff(coff);
    }

    return 0;
}
