/* -*-mb0-c-*- */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <time.h>		/* Used for ctime function */

/* Headers for working with COFF files */

#include "define.h"
#include "init.h"
#include <coff.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <bw/boio.h>

#define scan_syment(off, sym) _scan_syment(off, &sym)
#define scan_lineent(ptr, off, sym) _scan_lineent(ptr, off, &sym)

/* Global variables */

long	num_sections;	/* Number of section */
long	section_seek;   /* Used to seek to first section */

long	symptr;		/* File pointer to symbol table entries */
long	num_symbols;	/* Number of symbols */

char	*str_tab;	/* Pointer to start of string char. array */
long	str_length;	/* Length in bytes of string array */

char    infilename[80];
char    outfilename[80];

caddr_t glob_se, se_ptr;
struct syment g_se;

FILE	*fd;		/* COFF file descriptor */
int debugflag;
int f_example, f_dbg, f_path, f_head;

struct option_desc command_opt[]=
{
	{'o', "outfile", "%s", O_VALUE, &outfilename},
	{'D', "debug", "%d", O_VALUE, &debugflag},
	{'H', "headers", "", O_FLAG, (int *)&f_head},
	{'e', "example", "", O_FLAG, (int *)&f_example},
	{'d', "dbg", "", O_FLAG, (int *)&f_dbg},
	{'p', "path", "", O_FLAG, (int *)&f_path},
	{0, "", "", O_VALUE, (int *)NULL}
};

struct line_info
{
	int pc;
	int line;
	int l_file;
	int l_func;
} LINE_INFO;

#define MAXLINES 100000

struct line_info lif[MAXLINES];
struct line_info *lif_ptr;
int nof_lines=0;
int text_size;

void read_headers(void);
void read_strings(void);
void read_sections(void);
void read_symbols(void);
int search_sline(int adress, int lower, int upper);
int search_sym(int offset, int print);
void print_ae(char *);
void print_se(struct syment *se);

void _scan_syment(long off, struct syment *se)
{
	void *be_ptr = glob_se + symptr + off*SYMENT_SIZE;

	memcpy(se->n_name, (char *)be_ptr, N_SYMNMLEN);
	if (!se->n_zeroes) {
		se->n_offset = bw_be32_to_cpu(se->n_offset);
	}
	be_ptr = (char *)be_ptr + N_SYMNMLEN;

	se->n_value  = bw_be32_to_cpu(*(int32_t *)be_ptr);
	be_ptr = (int32_t *)be_ptr  + 1;
	se->n_scnum  = bw_be16_to_cpu(*(int16_t *)be_ptr);
	be_ptr = (int16_t *)be_ptr  + 1;
	se->n_type   = bw_be16_to_cpu(*(uint16_t *)be_ptr);
	be_ptr = (uint16_t *)be_ptr + 1;
	se->n_sclass =                *(uint8_t *)be_ptr;
	be_ptr = (uint8_t *)be_ptr  + 1;
	se->n_numaux =                *(uint8_t *)be_ptr;
	be_ptr = (uint8_t *)be_ptr  + 1;
}

void _scan_lineent(long ptr, long off, struct lineno *ln)
{
	void *be_ptr = glob_se + ptr + off*LINENO_SIZE;

	ln->l_paddr = bw_be32_to_cpu(*(int32_t *)be_ptr);
	be_ptr = (int32_t *)be_ptr  + 1;
	ln->l_lnno  = bw_be32_to_cpu(*(int32_t *)be_ptr);
	be_ptr = (int32_t *)be_ptr  + 1;
}

extern void eval_params(int argc, char *argv[]);

int
main(int argc,char *argv[])
{
	int pc;
	int index;
	int success=1;
	int fdesk;
	char s;
	
	eval_params(argc,argv);
	debugflag |= (f_example << S_EXA) | (f_dbg << S_DBG) | (f_head << S_HEAD);
	
	fd = fopen(infilename, "r");
	if (fd == 0) exit(0);

	fdesk=open(infilename, O_RDONLY);
	read_headers();
	glob_se = mmap(0, symptr+num_symbols*SYMENT_SIZE+1000, PROT_READ,
							 MAP_SHARED, fdesk, 0);
	close(fdesk);

	xif(S_DBG)
		read_strings();
	read_sections();
	xif(S_EXA)
		read_strings();
	read_symbols();
	xif(S_DBG)
		while(success!=EOF)
		{
			printf("pc: ");
			fflush(stdout);
			success=scanf("%i", &pc);
			if(success==0)
			{
				printf("Error\n");
				fgets(&s, 1, stdin);
			}
			else
			{
				if(pc > text_size)
					printf("pc too large!!!\n");
				else
				{
					index = search_sline(pc, 0, nof_lines-1);
					printf("pc# 0x%x, real=0x%x, source line=%d,in file ",
						   pc, lif[index].pc, lif[index].line);
					search_sym(lif[index].l_file, 1);      /* Name of File */
					if(lif[index].l_func!=0)
					{
						printf(", in function ");
						search_sym(lif[index].l_func, 1);
					}
					printf("\n");
				}
			}
		}
	exit(0);
}

int search_sline(int adress, int lower, int upper)
{
	int median;

	while(upper - lower > 1)
	{
		if(lif[median = (lower+upper)/2].pc <= adress)
		{
			lower = median;
		}
		else
		{
			upper = median - 1 ;
		}

	} /* while */

	if(lif[upper].pc > adress)
	{
		return lower;
	}
	else
	{
		return upper;
	}

} /* search_line */

int search_sym(int offset, int print)
{
	int i;
	
	scan_syment(offset, g_se);
	if(print)
	{
		if(g_se.n_zeroes)
		{
			for ( i = 0; i < N_SYMNMLEN; i++)
				if ((g_se.n_name[i] > 0x1f) && (g_se.n_name[i] < 0x7f))
					printf("%c", g_se.n_name[i]);
		}
		else
			printf("%s",&str_tab[g_se.n_offset-4]);
	}

	return 1;
} /* search_sym */

void
read_headers(void)
{
	struct filehdr  file_header;	/* File header structure */
	struct aouthdr	optional_header;/* Optional header structure */
	time_t t;

	coff_filehdr_fread(&file_header, 1, fd);

	xprintf(S_EXA)("FILE HEADER VALUES\n");
	xprintf(S_EXA)("f_magic  = 0%o\n", file_header.f_magic);
	if(file_header.f_magic!=SBP_MAGIC)
	{
		xprintf(S_EXA)("WARNING: WRONG MAGIC NUMBER, example may dump core!!!\n");
	}
	xprintf(S_EXA)("f_nscns  = %d\n",file_header.f_nscns);
	t = file_header.f_timdat;
	xprintf(S_EXA)("f_timdat = %s",ctime(&t));
	xprintf(S_EXA)("f_symptr = %d\n",file_header.f_symptr);
	xprintf(S_EXA)("f_nsyms  = %d\n",file_header.f_nsyms);
	xprintf(S_EXA)("f_opthdr = %d\n",file_header.f_opthdr);
	xprintf(S_EXA)("f_flags  = 0%o\n",file_header.f_flags);

	/* Save the global values */

	num_sections = file_header.f_nscns;
	num_symbols  = file_header.f_nsyms;
	symptr       = file_header.f_symptr;

	if (file_header.f_opthdr == AOUTHDR_SIZE)
	{
		coff_aouthdr_fread(&optional_header, 1, fd);

		xprintf(S_EXA)("OPTIONAL HEADER VALUES\n");
		xprintf(S_EXA)("magic      = 0%o\n", optional_header.magic);
		xprintf(S_EXA)("vstamp     = %d\n",optional_header.vstamp);
		xprintf(S_EXA)("tsize      = %d\n",optional_header.tsize);
		xprintf(S_EXA)("dsize      = %d\n",optional_header.dsize);
		xprintf(S_EXA)("bsize      = %d\n",optional_header.bsize);
		xprintf(S_EXA)("entry      = 0x%x\n",optional_header.entry);
		xprintf(S_EXA)("text_start = 0x%x\n",optional_header.text_start);
		xprintf(S_EXA)("data_start = 0x%x\n",optional_header.data_start);
	}
	else
	{
		fseek(fd,file_header.f_opthdr,SEEK_CUR);
	}

	/* File offset for first section headers */
	section_seek = FILEHDR_SIZE + file_header.f_opthdr;
}

void
read_sections(void)
{
	struct scnhdr	sh; /* Section header structure */
	struct reloc	re; /* Relocation entry structure */
	struct lineno	le; /* Line number entry structure */
	
	char	*raw_data;
	int i,j;
	int file_symdx=0, func_symdx=0;
	struct line_info *lif_ptr_func;
	
	lif_ptr=lif;
	
	for (i = 0; i < num_sections; i++)
	{
		fseek(fd,section_seek,0);

		coff_scnhdr_fread(&sh, 1, fd);

		section_seek += SCNHDR_SIZE;
		
		if(sh.s_flags==STYP_TEXT)
			text_size=sh.s_size/4;
		
		xprintf(S_EXA)("\n %s - SECTION HEADER - \n",sh.s_name);
		xprintf(S_EXA)("s_paddr   = 0x%x\n",sh.s_paddr);
		xprintf(S_EXA)("s_vaddr   = 0x%x\n",sh.s_vaddr);
		xprintf(S_EXA)("s_size    = %d\n",sh.s_size);
		xprintf(S_EXA)("s_scnptr  = %d\n",sh.s_scnptr);
		xprintf(S_EXA)("s_relptr  = %d\n",sh.s_relptr);
		xprintf(S_EXA)("s_lnnoptr = %d\n",sh.s_lnnoptr);
		xprintf(S_EXA)("s_nreloc  = %d\n",sh.s_nreloc);
		xprintf(S_EXA)("s_nlnno   = %d\n",sh.s_nlnno);
		xprintf(S_EXA)("s_flags   = 0x%x\n",sh.s_flags);

		/* for Section-Header Output */
		xprintf(S_HEAD)("%s\t: size %10d\n", sh.s_name, sh.s_size);
		
		/* Output raw data only for text and data sections */
		if (strcmp(sh.s_name,".bss") && (sh.s_flags!=STYP_BSS))
		{
			raw_data = (char *)malloc(sh.s_size);
			fseek(fd,sh.s_scnptr,0);
			fread(raw_data,sh.s_size,1,fd);
   
			xprintf(S_EXA)("RAW DATA\n");
			j=0;
			while(j<sh.s_size)
			{
				xprintf(S_EXA)("%02x ",(*raw_data&0xFF));
				raw_data++;
				j++;
				if (j%16==0) xprintf(S_EXA)("\n");
			}
			xprintf(S_EXA)("\n");
			free(raw_data-j);
		}

		if (sh.s_nreloc)
		{
			xprintf(S_EXA)("\nRELOCATION ENTRIES\n");
			fseek(fd,sh.s_relptr,0);
			j=0;
			while(j<sh.s_nreloc)
			{
				coff_reloc_fread(&re, 1, fd);

				xprintf(S_EXA)("r_vaddr = 0x%x",re.r_vaddr);
				xprintf(S_EXA)(" r_symndx = 0x%x",re.r_symndx);
				xprintf(S_EXA)(" r_offset = 0x%x\n",re.r_offset);
				j++;
			}
		}

		if (sh.s_nlnno)
		{
			xprintf(S_EXA)("\nLINE NUMBER ENTRIES \n");
			fseek(fd,sh.s_lnnoptr,0);
			j=0;
			while(j<sh.s_nlnno)
			{
				scan_lineent(sh.s_lnnoptr, j, le);
				if (le.l_lnno == 0)
				{
					xprintf(S_EXA)("function address 0x%x\n",le.l_symndx);

					scan_syment(le.l_symndx, g_se);
					
					xif(S_DBG)
					switch(g_se.n_sclass)
					{
					case C_FILE:
						file_symdx=le.l_symndx;
						/* printf("Function: 0x%x\n", file_symdx); */
						break;
					case C_EFCN:
						lif_ptr->l_func=0;
						lif_ptr_func=lif_ptr;
						while(--lif_ptr_func!=NULL && lif_ptr_func->l_func==0)
							lif_ptr_func->l_func=le.l_symndx;
						func_symdx=le.l_symndx;
						/* printf("Function: 0x%x\n", tmp_symdx); */
						break;
					default:
						break;
					}
				}
				else
				{
					xprintf(S_EXA)("line# %d at address 0x%x\n",le.l_lnno,le.l_paddr);
					lif_ptr->pc=le.l_paddr;
					lif_ptr->line=le.l_lnno;
					lif_ptr->l_file=file_symdx;
					lif_ptr->l_func=0;
					lif_ptr++;
					nof_lines++;
				}
				j++;
			}
		}

	}
}

void
read_strings(void)
{
	int	strings;
	char	*str_ptr;
	
	strings = symptr+SYMENT_SIZE*num_symbols;
	fseek(fd,strings,0);
	bw_fread_be32(&str_length,1,fd);
	if (str_length)
	{
		xprintf(S_EXA)("\nSTRING TABLE DUMP\n");
		str_length-=4;
		str_tab = (char *)malloc(str_length);
		fseek(fd,(strings+4),0);
		fread(str_tab,str_length,1,fd);
		str_ptr=str_tab;
		do
		{
			xprintf(S_EXA)("%s\n",str_ptr);
			while(*str_ptr++ !='\0');
		}
		while(str_ptr<(str_tab+str_length));
	}
}

void
read_symbols(void)
{
	struct syment	se;
	char  ae[AUXENT_SIZE];
	int	i;
	int	j;
	
	xprintf(S_EXA)("\nSYMBOL TABLE ENTRIES\n");
	i = 0;
	while (i  != num_symbols)
	{
		scan_syment(i, se);
		xprintf(S_EXA)("0x%05x ", i);
		print_se(&se); i++;
		for (j = 0; j < se.n_numaux; j++)
		{
			fread(ae,AUXENT_SIZE,1,fd);
			print_ae(ae);
			i++;
		}
	}
}

void
print_se(struct syment *se)
{
	int i;
	if (se->n_zeroes)
	{
		for ( i = 0; i < N_SYMNMLEN; i++)
		{
			if ((se->n_name[i] > 0x1f) && (se->n_name[i] < 0x7f))
			{
				xprintf(S_EXA)("%c",se->n_name[i]);
			}
			else
			{
				xprintf(S_EXA)(" ");
			}
		}
	}
	else
	{
		xprintf(S_EXA)("%s",&str_tab[se->n_offset-4]);
	}

	xprintf(S_EXA)(" n_scnum=%d ",se->n_scnum);

	switch (se->n_type & 0xf)
	{
		case T_CHAR:   xprintf(S_EXA)("T_CHAR"); break;
		case T_SHORT:  xprintf(S_EXA)("T_SHORT"); break;
		case T_INT:    xprintf(S_EXA)("T_INT"); break;
		case T_LONG:   xprintf(S_EXA)("T_LONG"); break;
		case T_FLOAT:  xprintf(S_EXA)("T_FLOAT"); break;
		case T_DOUBLE: xprintf(S_EXA)("T_DOUBLE"); break;
		case T_STRUCT: xprintf(S_EXA)("T_STRUCT"); break;
		case T_UNION:  xprintf(S_EXA)("T_UNION"); break;
		case T_ENUM:   xprintf(S_EXA)("T_ENUM"); break;
		case T_MOE:    xprintf(S_EXA)("T_MOE"); break;
		case T_UCHAR:  xprintf(S_EXA)("T_UCHAR"); break;
		case T_USHORT: xprintf(S_EXA)("T_USHORT"); break;
		case T_UINT:   xprintf(S_EXA)("T_UINT"); break;
		case T_ULONG:  xprintf(S_EXA)("T_ULONG"); break;
		default:       xprintf(S_EXA)("n_type=%d",se->n_type&0xf); break;
	}

	if (ISFCN(se->n_type))
	{
		xprintf(S_EXA)(",DT_FCN ");
	}
	else if (ISPTR(se->n_type))
	{
		xprintf(S_EXA)(",DT_PTR ");
	}
	else if (ISARY(se->n_type))
	{
		xprintf(S_EXA)(",DT_ARY ");
	}

	switch (se->n_sclass)
	{
		case C_NULL   : xprintf(S_EXA)(" C_NULL=0x%x",se->n_value); break;
		case C_STAT   : xprintf(S_EXA)(" C_STAT=0x%x",se->n_value); break;
		case C_EXT    : xprintf(S_EXA)(" C_EXT=0x%x",se->n_value); break;
		case C_AUTO   : xprintf(S_EXA)(" C_AUTO=0x%x",se->n_value); break;
		case C_FILE   : xprintf(S_EXA)(" C_FILE"); break;
		case C_STRTAG : xprintf(S_EXA)(" C_STRTAG"); break;
		case C_EOS    : xprintf(S_EXA)(" C_EOS=0x%x",se->n_value); break;
		case C_FCN    : xprintf(S_EXA)(" C_FCN=0x%x",se->n_value); break;
		case C_EFCN   : xprintf(S_EXA)(" C_EFCN=0x%x",se->n_value); break;
		default       : xprintf(S_EXA)(" n_sclass=0x%x",se->n_sclass); break;
	}

	xprintf(S_EXA)(" n_numaux=%d\n",se->n_numaux);
}

void
print_ae(char *ae)
{
	size_t i;
	xprintf(S_EXA)(" Aux. = ");
	for( i = 1; i <= AUXENT_SIZE; i++ )
	{
		xprintf(S_EXA)("%1x",(*ae>>4)&0xF);
		xprintf(S_EXA)("%1x ",(*ae&0xF));
		ae++;
	}

	xprintf(S_EXA)("\n");
}
