#ifndef _BW_BOIO_H
#define _BW_BOIO_H 1

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>

#include "bo.h"

static size_t
bw_fwrite_8(const void *ptr, size_t nmemb, FILE *stream)
{
    return fwrite(ptr, sizeof(char), nmemb, stream);
}

static size_t
bw_fread_8(void *ptr, size_t nmemb, FILE *stream)
{
    return fread(ptr, sizeof(char), nmemb, stream);
}

#define declare_fboio(__nbits, __endian)                                    \
static size_t                                                               \
bw_fwrite_##__endian##__nbits(const void *ptr, size_t nmemb, FILE *stream)  \
{                                                                           \
    uint##__nbits##_t memb;                                                 \
    const uint##__nbits##_t *memb_ptr = ptr;                                \
    size_t i, tret = 0, ret;                                                \
                                                                            \
    if (!nmemb)                                                             \
        return 0;                                                           \
                                                                            \
    for (i = 0; i < nmemb; i++) {                                           \
        memb = bw_cpu_to_##__endian##__nbits(memb_ptr[i]);                  \
        ret = fwrite(&memb, sizeof(memb), 1, stream);                       \
        if (!ret)                                                           \
            return tret;                                                    \
        tret += ret;                                                        \
    }                                                                       \
                                                                            \
    return tret;                                                            \
}                                                                           \
                                                                            \
static size_t                                                               \
bw_fread_##__endian##__nbits(void *ptr, size_t nmemb, FILE *stream)         \
{                                                                           \
    uint##__nbits##_t memb;                                                 \
    uint##__nbits##_t *memb_ptr = ptr;                                      \
    size_t i, tret = 0, ret;                                                \
                                                                            \
    if (!nmemb)                                                             \
        return 0;                                                           \
                                                                            \
    for (i = 0; i < nmemb; i++) {                                           \
        do {                                                                \
            ret = fread(&memb, sizeof(memb), 1, stream);                    \
        } while (!(ret || ferror(stream) || feof(stream)));                 \
        if (!ret)                                                           \
            return tret;                                                    \
        memb_ptr[i] = bw_cpu_to_##__endian##__nbits(memb);                  \
        tret += ret;                                                        \
    }                                                                       \
                                                                            \
    return tret;                                                            \
}


declare_fboio(16, be)
declare_fboio(16, le)

declare_fboio(32, be)
declare_fboio(32, le)

#if defined __GNUC__ && __GNUC__ >= 2

declare_fboio(64, be)
declare_fboio(64, le)

#endif

#undef declare_fboio

#endif /* _BW_BOIO_H */
