#ifndef _BW_ST_H
#define _BW_ST_H 1

#define _GNU_SOURCE
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdbool.h>
#include <math.h>
#include <errno.h>

#include "bw.h"

struct bw_st;

typedef void (*bw_st_free_fn_t)(struct bw_st *, void *);

typedef struct bw_st {
    struct bw_st *x;
} bw_st_t;

static  int
bw_st_init(bw_st_t *ste)
{
    if (!ste)
        return EINVAL;

    ste->x = NULL;

    return ESUCCESS;
}

#define BW_ST_INIT() { .x = NULL }

static bw_st_t *
bw_st_next(bw_st_t *ste)
{
    if (!ste) {
        errno = EINVAL;
        return NULL;
    }

    return ste->x;
}

static int
bw_st_free(bw_st_t **st, bw_st_free_fn_t free_fn, void *ctx)
{
    bw_st_t *e;

    if (!st)
        return EINVAL;

    if (free_fn) {
        while (*st) {
            e = (*st)->x;
            (*st)->x = NULL;
            free_fn(*st, ctx);
            *st = e;
        }
    } else {
        while (*st) {
            e = (*st)->x;
            (*st)->x = NULL;
            *st = e;
        }
    }

    return ESUCCESS;
}

static int
bw_st_del(bw_st_t **st, bw_st_t *e)
{
    if (!st || !e)
        return EINVAL;

    while (*st) {
        if (*st == e) {
            *st = e->x;
            e->x = NULL;
            break;
        }
        st = &(*st)->x;
    }

    return ESUCCESS;
}

static int
bw_st_push(bw_st_t **st, bw_st_t *e)
{
    if (!st || !e)
        return EINVAL;

    e->x = *st;
    *st = e;

    return ESUCCESS;
}

static bw_st_t *
bw_st_pop(bw_st_t **st)
{
    bw_st_t *e = NULL;

    if (!st) {
        errno = EINVAL;
        return NULL;
    }

    if (*st) {
        e = *st;
        (*st) = e->x;
        e->x = NULL;
    }

    return e;
}

/* insert `e` after `*st`
 */
static int
bw_st_ins(bw_st_t **st, bw_st_t *e)
{
    if (!st || !e)
        return EINVAL;

    if (*st) {
        e->x = (*st)->x;
        (*st)->x = e;
    } else {
        e->x = NULL;
        *st = e;
    }

    return ESUCCESS;
}

/* insert `e` after `*st` and advance `*st` to `e`
 * good for tail of queue
 */
static int
bw_st_insa(bw_st_t **st, bw_st_t *e)
{
    if (!st || !e)
        return EINVAL;

    if (*st) {
        e->x = (*st)->x;
        (*st)->x = e;
    } else
        e->x = NULL;
    *st = e;

    return ESUCCESS;
}

#define bw_st_forall(__e, __st, __m, ...)                       \
({                                                              \
    bw_st_t *__st_fa_st = (__st);                               \
    while (__st_fa_st) {                                        \
        __e = container_of(__st_fa_st, typeof(*(__e)), __m);    \
        (__VA_ARGS__);                                          \
        __st_fa_st = bw_st_next(__st_fa_st);                    \
    }                                                           \
});

#endif /* _BW_ST_H */
