/* -*-mbo-c-*- */
/* FILE:        coff.h
   DESCRIPTION: defines for coff.c
   AUTHOR:      Stefan Franziskus
   DATE:        DEC 14, 1993
*/

#ifndef _COFF_H
#define _COFF_H 1

#include <stdint.h>
#include <inttypes.h>

/* the SB-PRAM coff f_magic */
#define SBP_MAGIC        UINT16_C(0x0500)

#define STYP_REG         UINT16_C(0x0000)       /* "regular": allocated, reocated, loaded */
#define STYP_DSECT       UINT16_C(0x0001)       /* "dummy":  relocated only*/
#define STYP_NOLOAD      UINT16_C(0x0002)       /* "noload": allocated, relocated, not loaded */
#define STYP_GROUP       UINT16_C(0x0004)       /* "grouped": formed of input sections */
#define STYP_PAD         UINT16_C(0x0008)       /* "padding": not allocated, not relocated, loaded */
#define STYP_COPY        UINT16_C(0x0010)       /* "copy": for decision function used by field update;  not allocated, not relocated,
                                                                             loaded; reloc & lineno entries processed normally */
#define STYP_TEXT        UINT16_C(0x0020)       /* section contains text only */
#define S_SHRSEG         UINT16_C(0x0020)       /* In 3b Update files (output of ogen), sections which appear in SHARED segments of the Pfile
                                                                             will have the S_SHRSEG flag set by ogen, to inform dufr that updating 1 copy of the proc. will
                                                                             update all process invocations. */
#define STYP_DATA        UINT16_C(0x0040)       /* section contains data only */
#define STYP_BSS         UINT16_C(0x0080)       /* section contains bss only */
#define S_NEWFCN         UINT16_C(0x0100)       /* In a minimal file or an update file, a new function (as compared with a replaced function) */

#define STYP_INFO        UINT16_C(0x0200)       /* comment: not allocated not relocated, not loaded */
#define STYP_OVER        UINT16_C(0x0400)       /* overlay: relocated not allocated or loaded */
#define STYP_LIB         UINT16_C(0x0800)       /* for .lib: same as INFO */
#define STYP_MERGE       UINT16_C(0x2000)       /* merge section -- combines with text, data or bss sections only */
#define STYP_REVERSE_PAD UINT16_C(0x4000)       /* section will be padded with no-op instructions wherever padding is necessary and there is a

                                                                             word of contiguous bytes
                                                                             beginning on a word boundary. */

#define STYP_LIT        UINT16_C(0x8020)        /* Literal data (like STYP_TEXT) */


struct filehdr
{
	uint16_t   f_magic;      /* magic number */
	uint16_t   f_nscns;      /* number of sections */
	int32_t    f_timdat;     /* time & date stamp */
	int32_t    f_symptr;     /* file pointer to symtab */
	int32_t    f_nsyms;      /* number of symtab entries */
	uint16_t   f_opthdr;     /* sizeof(optional hdr) */
	uint16_t   f_flags;      /* flags */
};
#define FILEHDR_SIZE ((size_t)20u)

/*
 * Bits for f_flags:
 */
#define F_RELFLG        UINT16_C(0x0001)    /* relocation info stripped from file */
#define F_EXEC          UINT16_C(0x0002)    /* file is executable (no unresolved external references) */
#define F_LNNO          UINT16_C(0x0004)    /* line numbers stripped from file */
#define F_LSYMS         UINT16_C(0x0008)    /* local symbols stripped from file */
#define F_AR16WR        UINT16_C(0x0080)    /* file is 16-bit little-endian */
#define F_AR32WR        UINT16_C(0x0100)    /* file is 32-bit little-endian */
#define F_AR32W         UINT16_C(0x0200)    /* file is 32-bit big-endian */
#define F_DYNLOAD       UINT16_C(0x1000)    /* rs/6000 aix: dynamically loadable w/imports & exports */
#define F_SHROBJ        UINT16_C(0x2000)    /* rs/6000 aix: file is a shared object */

struct aouthdr
{
	int16_t         magic;        /* magic number */
	int16_t         vstamp;       /* version stamp */
	int32_t         tsize;        /* text size in bytes */
	int32_t         dsize;        /* initialized data size */
	int32_t         bsize;        /* uninitialized data size */
	int32_t         entry;        /* entry point */
	int32_t         text_start;   /* base of text used for this file */
	int32_t         data_start;   /* base of data used for this file */
};
#define AOUTHDR_SIZE ((size_t)28u)

#define S_SCNNMLEN 8
struct scnhdr
{
	char        s_name[S_SCNNMLEN + 1];    /* section name */
	uint32_t    s_paddr;      /* physical address */
	uint32_t    s_vaddr;      /* virtual address */
	int32_t     s_size;       /* section size */
	int32_t     s_scnptr;     /* file ptr to raw data for section */
	int32_t     s_relptr;     /* file ptr to relocation */
	int32_t     s_lnnoptr;    /* file ptr to line numbers */
	uint16_t    s_nreloc;     /* number of relocation entries */
	uint16_t    s_nlnno;      /* number of line number entries */
	int32_t     s_flags;      /* type and content flags */
};
#define SCNHDR_SIZE ((size_t)40u)

/********************** STORAGE CLASSES **********************/

/* This used to be defined as -1, but now n_sclass is unsigned.  */
#define C_EFCN          UINT8_C(255)     /* physical end of function     */
#define C_NULL          UINT8_C(  0)
#define C_AUTO          UINT8_C(  1)     /* automatic variable           */
#define C_EXT           UINT8_C(  2)     /* external symbol              */
#define C_STAT          UINT8_C(  3)     /* static                       */
#define C_REG           UINT8_C(  4)     /* register variable            */
#define C_EXTDEF        UINT8_C(  5)     /* external definition          */
#define C_LABEL         UINT8_C(  6)     /* label                        */
#define C_ULABEL        UINT8_C(  7)     /* undefined label              */
#define C_MOS           UINT8_C(  8)     /* member of structure          */
#define C_ARG           UINT8_C(  9)     /* function argument            */
#define C_STRTAG        UINT8_C( 10)     /* structure tag                */
#define C_MOU           UINT8_C( 11)     /* member of union              */
#define C_UNTAG         UINT8_C( 12)     /* union tag                    */
#define C_TPDEF         UINT8_C( 13)     /* type definition              */
#define C_USTATIC       UINT8_C( 14)     /* undefined static             */
#define C_ENTAG         UINT8_C( 15)     /* enumeration tag              */
#define C_MOE           UINT8_C( 16)     /* member of enumeration        */
#define C_REGPARM       UINT8_C( 17)     /* register parameter           */
#define C_FIELD         UINT8_C( 18)     /* bit field                    */
#define C_AUTOARG       UINT8_C( 19)     /* auto argument                */
#define C_LASTENT       UINT8_C( 20)     /* dummy entry (end of block)   */
#define C_BLOCK         UINT8_C(100)     /* ".bb" or ".eb"               */
#define C_FCN           UINT8_C(101)     /* ".bf" or ".ef"               */
#define C_EOS           UINT8_C(102)     /* end of structure             */
#define C_FILE          UINT8_C(103)     /* file name                    */
#define C_LINE          UINT8_C(104)     /* line # reformatted as symbol table entry */
#define C_ALIAS         UINT8_C(105)     /* duplicate tag                */
#define C_HIDDEN        UINT8_C(106)     /* ext symbol in dmert public lib */

struct reloc
{
	int32_t          r_vaddr;      /* address of reference */
	int32_t          r_symndx;     /* index into symbol table */
	uint32_t         r_offset;
	uint16_t         r_type;       /* relocation type */
	int16_t          r_stuff;
};
#define RELOC_SIZE ((size_t)16u)

/* Funny relocation types and other stuff */

#define R_13REL    UINT16_C(0x01)
#define R_13ABS    UINT16_C(0x02)
#define R_19REL    UINT16_C(0x04)
#define R_19ABS    UINT16_C(0x08)
#define R_32DAT    UINT16_C(0x10)
#define R_GETHI    UINT16_C(0x20)

#define REL_MASK13 UINT32_C(0xFFFFE000)
#define SIGN_13    UINT32_C(0x00002000)
#define REL_MASK19 UINT32_C(0xFFF80000)
#define SIGN_19    UINT32_C(0x00080000)
#define REL_MASK32 UINT32_C(0x00000000)

#define N_SYMNMLEN 8

struct syment
{
	union
	{
		char _n_name[N_SYMNMLEN + 1];           /* Symbol name */
		struct                           /* if _n_name[0-3] == 0 */
		{
			int32_t _n_zeroes;           /* then _n_name [4-7] is an */
			int32_t _n_offset;           /* offset into string table */
		} _n_n;
	} _n;
	int32_t           n_value;           /* value of symbol */
	int16_t           n_scnum;           /* section number */
	uint16_t          n_type;            /* type and derived type */
	uint8_t           n_sclass;          /* storage class */
	uint8_t           n_numaux;          /* number of aux. entries */
};
#define n_name    _n._n_name
#define n_zeroes  _n._n_n._n_zeroes
#define n_offset  _n._n_n._n_offset
#define SYMENT_SIZE ((size_t)18u)

union auxent {
	struct {
		uint32_t x_tagndx;  /* str, un, or enum tag indx */
		union {
			struct {
				uint16_t x_lnno;    /* declaration line number */
				uint16_t x_size;    /* str/union/array size */
			} x_lnsz;
			uint32_t x_fsize;   /* size of function */
		} x_misc;
		union {
			struct {    /* if ISFCN, tag, or .bb */
				uint32_t x_lnnoptr; /* ptr to fcn line # */
				uint32_t x_endndx;  /* entry ndx past block end */
			} x_fcn;
			struct {    /* if ISARY, up to 4 dimen. */
				uint16_t x_dimen[4];
			} x_ary;
		} x_fcnary;
		uint16_t x_tvndx;   /* tv index */
	} x_sym;

	union {
		char x_fname[14];
		struct {
			uint32_t x_zeroes;
			uint32_t x_offset;
		} x_n;
	} x_file;

	struct {
		uint32_t x_scnlen;  /* section length */
		uint16_t x_nreloc;  /* # relocation entries */
		uint16_t x_nlinno;  /* # line numbers */
	} x_scn;

	struct {
		uint32_t x_tvfill;      /* tv fill value */
		uint16_t x_tvlen;       /* length of .tv */
		uint16_t x_tvran[2];    /* tv range */
	} x_tv; /* info about .tv section (in auxent of symbol .tv)) */

};
#define AUXENT_SIZE ((size_t)18u)

#define N_BTMASK    (017)
#define N_TMASK     (060)
#define N_BTSHFT      (4)
#define N_TSHIFT      (2)

/*
 * Type of a symbol, in low N bits of the word
 */
#define T_NULL       0
#define T_VOID       1      /* function argument (only used by compiler) */
#define T_CHAR       2      /* character */
#define T_SHORT      3      /* short integer */
#define T_INT        4      /* integer */
#define T_LONG       5      /* long integer */
#define T_FLOAT      6      /* floating point */
#define T_DOUBLE     7      /* double word */
#define T_STRUCT     8      /* structure */
#define T_UNION      9      /* union */
#define T_ENUM      10      /* enumeration */
#define T_MOE       11      /* member of enumeration*/
#define T_UCHAR     12      /* unsigned character */
#define T_USHORT    13      /* unsigned short */
#define T_UINT      14      /* unsigned integer */
#define T_ULONG     15      /* unsigned long */
#define T_LNGDBL    16      /* long double */

/*
 * derived types, in n_type
*/
#define DT_NON  (0)     /* no derived type */
#define DT_PTR  (1)     /* pointer */
#define DT_FCN  (2)     /* function */
#define DT_ARY  (3)     /* array */

#define BTYPE(x)    ((x) & N_BTMASK)

#define ISPTR(x)    (((x) & N_TMASK) == (DT_PTR << N_BTSHFT))
#define ISFCN(x)    (((x) & N_TMASK) == (DT_FCN << N_BTSHFT))
#define ISARY(x)    (((x) & N_TMASK) == (DT_ARY << N_BTSHFT))
#define ISTAG(x)    ((x)==C_STRTAG||(x)==C_UNTAG||(x)==C_ENTAG)

struct lineno
{
	union
	{
		int32_t _l_symndx;    /* function name symbol index, iff l_lnno == 0*/
		int32_t _l_paddr;     /* (physical) address of line number    */
	} _l; /* l_addr */
	int32_t l_lnno;          /* line number          */
};
#define l_symndx _l._l_symndx
#define l_paddr  _l._l_paddr
#define LINENO_SIZE  ((size_t)8u)

/* I/O helper functions */
extern size_t coff_filehdr_fread(struct filehdr *filehdr, size_t nmemb, FILE *f);
extern size_t coff_aouthdr_fread(struct aouthdr *aouthdr, size_t nmemb, FILE *f);
extern size_t coff_scnhdr_fread(struct scnhdr *scnhdr, size_t nmemb, FILE *f);
extern size_t coff_reloc_fread(struct reloc *reloc, size_t nmemb, FILE *f);
extern size_t coff_syment_fread(struct syment *syment, size_t nmemb, FILE *f);
extern size_t coff_lineno_fread(struct lineno *lineno, size_t nmemb, FILE *f);

extern size_t coff_filehdr_fwrite(const struct filehdr *filehdr, size_t nmemb, FILE *f);
extern size_t coff_aouthdr_fwrite(const struct aouthdr *aouthdr, size_t nmemb, FILE *f);
extern size_t coff_scnhdr_fwrite(const struct scnhdr *scnhdr, size_t nmemb, FILE *f);
extern size_t coff_reloc_fwrite(const struct reloc *reloc, size_t nmemb, FILE *f);
extern size_t coff_syment_fwrite(const struct syment *syment, size_t nmemb, FILE *f);
extern size_t coff_lineno_fwrite(const struct lineno *lineno, size_t nmemb, FILE *f);

#endif /* _COFF_H */
