/* -*-mb0-c-*- */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>
#include <errno.h>
#include <error.h>
#include <stdint.h>
#include <inttypes.h>

#include "loader.h"
#include "define.h"
#include "init.h"
#include "loaderinit.h"

#define strget() \
({ \
	if((ld_entry = strtok(NULL, "\t \n")) == NULL) \
	{ \
		if((success=fgets(string, 512, configfile)) == NULL) \
			break;\
		ld_entry = strtok(string , "\t \n"); \
	} \
})

char *infilename="default", *outfilename;   /* names of files */
int debugflag=0;
int32_t textstart=0, datastart=0;

#ifdef HOST
int message_key;
#endif

unsigned options;

#ifdef EXPORT

char loaderrc[4096]="/local/fork/fork/lib/ldrc.fork";

#else /* if stefran */

char loaderrc[4096]=".ldrc";

#endif /* EXPORT */

struct option_desc command_opt[]=
{
	{'h', "help",         O_HELP, {},        ""},
	{'o', "outfile",      I_STRING(outfilename),
		  "Name of the output file"},
	{'D', "debugflags",   I_INT(&debugflag),
		  "debugflags"},
	{'t', "textstart",    I_INT(&textstart),
		  "Start of the text-section in the program memory"},
	{'d', "datastart",    I_INT(&datastart),
		  "Start of the data-sections in the global memory"},
	{'l', "loaderrc",     I_STRINGBUF(loaderrc),
		  "name and path of loaderrc"},
	{'L', "ldvalue",      I_FUNCTION(ldvalue_set),
		  "Set a new value in lddata"},
	{'F', "forkdebug",    I_FLAG(OPT_FORKDEBUG),
		  "Omit directives for the Fork Simulator/Debugger"},
	{'a', "args",    I_FLAG(OPT_ARGS),
		  "Realy emit \".args\" section"},
	{'2', "cod2",    I_FLAG(OPT_COD2),
		  "Emit a .cod2 version (with .SIZE directives)"},
	{' ', "pipe",    I_FLAG(OPT_PIPE),
		  "Emit to stdout"},
#ifdef HOST
	{' ', "host-com",     I_FLAG(OPT_HOST_COM),
		  "Communicate with the SB-PRAM-Host"},
	{' ', "host-msg-key", I_INT(&message_key),
		  "Descriptor for the message queue"},
#endif
	{'V', "version",      I_FUNCTION(print_version),
		  "Print the current version"},
	{' ', "sysv-messages", I_FLAG(OPT_SYSV_MSG),
		  "Use System V Messages instead of shared Memory"},
};

int command_opt_len = (sizeof(command_opt)/sizeof(*command_opt));

struct environ_desc environ_opt[]=
{
	{"SBP_LOADERRC",      loaderrc},
};

int environ_opt_len = (sizeof(environ_opt)/sizeof(*environ_opt));

void free_sct(struct section *sct)
{
	if(sct->next!=NULL)
		free_sct(sct->next);
	else
		free(sct);
}

void get_mem_values(void)
{
	char *ld_entry;
	char *success;
	uint32_t org=UINT32_C(0);
	uint32_t orgflag=UINT32_C(0);
	int mem_num=0;
	char string[512];
	struct lditem *ld_ptr;

	ld_ptr = ld_data;
	if(sct!=NULL)
		free_sct(sct);

	sct=calloc(1, sizeof(struct section));
	sct->next=NULL;
	sct_beg=sct;
	strncpy(sct->head.s_name,".text", S_SCNNMLEN);
	sct->head.s_name[S_SCNNMLEN] = '\0';
	
	success=fgets(string, 512, configfile);
	ld_entry = strtok(string , "\t \n");
	while(success!=NULL && ld_entry!=NULL)
	{
		if(!strcmp(ld_entry, "sctdata"))
		{
			strget();
			while(strcmp(ld_entry, "{") && success!=NULL) strget();
			strget();
			while(strcmp(ld_entry, "}") && success!=NULL)
			{
				if(!strcmp(ld_entry, "org"))
				{
					strget();
					if(!strcmp(ld_entry, "new"))
						orgflag=UINT32_C(0xffffffff);
					else
						if(!strcmp(ld_entry, "*"))
							orgflag=UINT32_C(0);
						else
						{
							org=(uint32_t)strtoul(ld_entry,NULL,0);
							orgflag=UINT32_C(1);
						}
					strget();
				}
				
				if(success && strlen(ld_entry)>0)
				{
					sct->next=calloc(1, sizeof(struct section));
					sct=sct->next;
					sct->next=NULL;
					strncpy(sct->head.s_name, ld_entry, S_SCNNMLEN);
					sct->head.s_name[S_SCNNMLEN] = '\0';
					sct->m_num = mem_num;
					sct->head.s_paddr=orgflag;
					sct->head.s_vaddr=org;
					sct->head.s_size=0;
					strget();
					if(!strcmp(ld_entry,"s") || !strcmp(ld_entry,"S"))
						sct->sbp_flags=SBP_SHARED;
					if(!strcmp(ld_entry,"p") || !strcmp(ld_entry,"P"))
						sct->sbp_flags=SBP_PRIVATE;
					if(orgflag==UINT32_C(0xffffffff))
						orgflag=0;
				}
				strget();
			}
			mem_num++;
		}
		if(!strcmp(ld_entry, "lddata"))
		{
			strget();
			while(strcmp(ld_entry, "{") && success!=NULL) strget();
			strget();
			while(strcmp(ld_entry, "}") && success!=NULL)
			{
				ld_ptr->name = strdup(ld_entry);
				strget();
				sscanf(ld_entry, "%i", &ld_ptr->value);
				strget();
				ld_ptr++->type = ld_entry[0];      /* error in file ? */
				strget();
			}
		}
		strget();
	}
}


void usage(const char *format, ...)
{
    int i;
    va_list arg;
    struct option_desc *co;
    
	fprintf(stderr, "SB-PRAM(TM) coff(TM)-loader\n\nusage:\n");
    if(format)
    {
        va_start(arg, format);
        vfprintf(stderr, format, arg);
        va_end(arg);
    }
    else
    {
        for(co=command_opt, i=0; i<command_opt_len; i++, co++)
        {
            fprintf(stderr, "%c  %c%c --%-15s  %s\n",
                    (co->type==O_FLAG && !co->val.flags)? 'd': ' ',
                    co->short_opt==' '? ' ': '-',
                    co->short_opt, co->long_opt, co->comment);
        }
    }
	i=0;
	fprintf(stderr, "\nCurrent ldvalues for lddata are:\n");
		while(i<LD_DATA_LEN && ld_data[i].name != NULL)
		{
			fprintf(stderr, "%10s: %8d\n", ld_data[i].name, ld_data[i].value);
			i++;
		}
	
    exit(format? 1: 0);
}

int first_run = 1;

int eval_params(int argc, char *argv[])
{ 
    int par, i;
    char *arg;
    int arg_shift;
    
    if(first_run)
    {
        options = 0;
        first_run = 0;
    }
        
    for(par=1; par<argc; par++)
    {
        arg = argv[par + 1];
        arg_shift = 1;

        if(argv[par][0] != '-') break;
        
        if(argv[par][1] == '-')
        {
            /* break on "--" */
            if (!argv[par][2]) {
                par++;
                break;
            }

            for(i=0; i<command_opt_len; i++) {
                if (!strncmp(argv[par] + 2, command_opt[i].long_opt, strlen(command_opt[i].long_opt))) {
                    if (argv[par][2 + strlen(command_opt[i].long_opt)] == '=') {
                        arg = &argv[par][3 + strlen(command_opt[i].long_opt)];
                        arg_shift = 0;
                        break;
                    } else if (!argv[par][2 + strlen(command_opt[i].long_opt)]) {
                        break;
                    }
                }
            }
        }
        else
        {
            for(i=0; i<command_opt_len; i++)
                if(argv[par][1]==command_opt[i].short_opt)
                    break;
        }
        if(i>=command_opt_len)
            usage("unknown option '%s'\n", argv[par]);
        else
        {
			switch(command_opt[i].type)
			{
			case O_INT:
				par += arg_shift;
				if(1!=sscanf(arg, "%i",
							 command_opt[i].val.int_val))
					usage("Option --%s expects a number\n",
						  command_opt[i].long_opt);
				break;
			case O_STRING:
				par += arg_shift;
				*command_opt[i].val.string_val = strdup(arg);
				break;
			case O_STRINGBUF:
				par += arg_shift;
				strcpy(command_opt[i].val.string_buf, arg);
				break;
			case O_FLAG:
				if(!command_opt[i].val.flags)
					usage("Option --%s disabled\n",
						  command_opt[i].long_opt);
				else if(command_opt[i].val.flags > 0)
					options |= command_opt[i].val.flags;
				else
					options &= command_opt[i].val.flags;
				break;
			case O_FUNCTION:
				par+=(command_opt[i].val.function)(argc - par - 1, &argv[par + 1]);
				break;
			case O_IGNORE:
				++par;
				break;
			case O_HELP:
				usage(NULL);
				break;
            }
        }
    }
	
	if (outfilename == NULL)
	{
		if (!(options & OPT_PIPE))
			outfilename = "a.cod";
	}
	else if (!strstr(outfilename, ".cod"))
	{
		char *new = malloc(strlen(outfilename) + strlen(".cod") + 1);
		sprintf(new, "%s.cod", outfilename);
		free(outfilename);
		outfilename = new;
	}
	
	if(par<argc)
		infilename = strdup(argv[par]);   /* !!! Noch Error ausgeben */
	
	sbp_argv=malloc((argc-par+2)*sizeof(char *));
	sbp_argc=argc-par;
	i=0;
	if (debugflag==1)
	{
		fprintf(stderr, "sbp_argc: %d\n", sbp_argc);
	}
	while(par<argc)
	{
		/* probably: strlen() + terminating nil + argv pointer */
		sbp_arglen+=strlen(argv[par])+2;
		sbp_argv[i]=argv[par++];
		if (debugflag==1)
		{
			fprintf(stderr, "sbp_arg[%d]: %s\n", i, sbp_argv[i]);
		}
		i++;
	}
	if (debugflag==1)
	{
		fprintf(stderr, "sbp_arglen: %d\n", sbp_arglen);
	}

	return par;
}

void get_environment(void)
{
    int i;
    char *env;
    
    for(i=0;i<environ_opt_len;i++)
    {
        if((env = getenv(environ_opt[i].env_name)) != NULL)
            strcpy(environ_opt[i].env_value, env);
    }
}

void syserr(char *msg)
{
  fprintf(stderr, "ERROR: %s (%d", msg, errno);
  if(errno >0)
	fprintf(stderr, ": %s)\n", strerror(errno));
  else
	fprintf(stderr, ")\n");
  exit(1);
}

