/* -*-mb0-c-*- */
/* 
  FILE:        loader.c
  DESCRIPTION: loader for coff-files
  AUTHOR:      Stefan Franziskus
  DATE:        JAN 15, 1993
*/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<sys/wait.h>
#include <unistd.h>
#include <stdint.h>
#include <inttypes.h>

#include <coff.h>

#include <bw/boio.h>

#include "define.h"
#include "loader.h"
#include "init.h"
#include "loaderinit.h"

#ifdef HOST
#ifndef LOADER
#define LOADER
#endif
#include "../host/shmem_com.h"
#include "../host/communication.h"
extern int dev_prog, dev_pram;
extern char *section_mem[];
extern int _mptr_, *GlobalMem;
#endif

char version[]="Version 0.8 (c) 1994 by stefran@cs.uni-sb.de";

FILE *infile, *dumpfile, *configfile;     /* corresp. file descr */

int quietflag;

struct section *sct;           /* Sectionliste */
struct section *sct_beg;       /* Pointer auf Anfang der Sectionliste */
struct syment *symtab;
static struct filehdr file_head;
static struct aouthdr aout_head;
static uint32_t v_addr[10];

struct lditem ld_data[LD_DATA_LEN];

char **sbp_argv;
/* sbp_argc:   at least the last NULL pointer,  */
/* sbp_arglen: at least the argc + 'reservepointer' + last NULL pointer, */
int sbp_argc=1, sbp_arglen=3;
unsigned options;

int print_version(int ai, char **av)
{
	fprintf(stderr, "%s\n", version);
	exit(0);
	return 0;
}

int ldvalue_set(int num, char *arg[])
{
	int i=0;
	
/*	if(num<2)
		errexit("Wrong Arguments"); */
	
	while(i<LD_DATA_LEN && ld_data[i].name != NULL)
	{
		if(!strcmp(ld_data[i].name, arg[0]))
		{
			sscanf(arg[1], "%i", &ld_data[i].value);
			return(2);
		}
		i++;
	}

	/* name not found */
	fprintf(stderr, "Unknown Argument to -L option\n");
	exit(1);
	return(0);
}

#define OP_GET 1
#define OP_PUT 2
#define get_size(name) lddata_size(OP_GET, name, 0)
#define put_size(name, val) lddata_size(OP_PUT, name, val)
inline int lddata_size(int op, char *name, int value)
{
	int i=0;
	while(i<LD_DATA_LEN && ld_data[i].name != NULL)
	{
		if(!strcmp(ld_data[i].name, name))
		{
			if(op==OP_GET)
			{
				return(ld_data[i].value);
			}
			else if(op==OP_PUT)
			{
				ld_data[i].value = value;
				return(0);
			}
		}
		i++;
	}
	return(-1);
}

inline int get_typesize(char type)
{
	int i=0, size=0;
	while(i<LD_DATA_LEN && ld_data[i].name != NULL)
	{
		if(ld_data[i].type == type)
			size += ld_data[i].value;
		i++;
	}
	return(size);
}

inline int get_section_size(int type)
{
	int size=0;
	
	sct=sct_beg;
	
	while(sct!=NULL)
	{
		if(type == SBP_SHARED || type == SBP_PRIVATE)
		{
			if(sct->sbp_flags == type)
				size += sct->head.s_size;
		}
		else if(sct->head.s_flags == STYP_TEXT)
			size += sct->head.s_size;
		sct = sct->next;
	}
	return(size);
}

void scan_coff_file(void)
{
	int32_t i=0,j=0,dummy;

	fseek(infile, 0, SEEK_SET);
	coff_filehdr_fread(&file_head, 1, infile);

	if(file_head.f_magic != SBP_MAGIC)
	{
		fprintf(stderr, "Error, inputfile is not a coff file %s\n", infilename);
		exit(1);
	}

	if(!(file_head.f_flags & F_EXEC))
	{
		fprintf(stderr, "Error, inputfile is not executable %s\n", infilename);
		exit(1);
	}

	coff_aouthdr_fread(&aout_head, 1, infile);

	if(debugflag==1)
	{
	    fprintf(stderr, "f_magic      %04x\n",file_head.f_magic);
	    fprintf(stderr, "f_nscns      %04x\n",file_head.f_nscns);
	    fprintf(stderr, "f_timdat %08x\n",file_head.f_timdat);
	    fprintf(stderr, "f_symptr %08x\n",file_head.f_symptr);
	    fprintf(stderr, "f_nsyms  %08x\n",file_head.f_nsyms);
	    fprintf(stderr, "f_opthdr     %04x\n",file_head.f_opthdr);
	    fprintf(stderr, "f_flags      %04x\n",file_head.f_flags);
	    fprintf(stderr, "\n");
	
	    fprintf(stderr, "magic       %04x\n",aout_head.magic);
	    fprintf(stderr, "vstamp      %04x\n",aout_head.vstamp);
	    fprintf(stderr, "tsize   %08x\n",aout_head.tsize);
	    fprintf(stderr, "dsize   %08x\n",aout_head.dsize);
	    fprintf(stderr, "bsize   %08x\n",aout_head.bsize);
	    fprintf(stderr, "entry   %08x\n",aout_head.entry);
	    fprintf(stderr, "text_st %08x\n",aout_head.text_start);
	    fprintf(stderr, "data_st %08x\n",aout_head.data_start);
	    fprintf(stderr, "\n");
	}
	
	sct=sct_beg;
	
	while(sct!=NULL)
	{
		fseek(infile, FILEHDR_SIZE + file_head.f_opthdr, SEEK_SET);

		for(i=0;i<file_head.f_nscns;++i)
		{
			struct scnhdr scn;

			coff_scnhdr_fread(&scn, 1, infile);

			if (!strcmp(scn.s_name,sct->head.s_name))
			{
				/* skip s_paddr and s_vaddr */
				sct->head.s_size    = scn.s_size / 4;
				sct->head.s_scnptr  = scn.s_scnptr;
				sct->head.s_relptr  = scn.s_relptr;
				sct->head.s_lnnoptr = scn.s_lnnoptr;
				sct->head.s_nreloc  = scn.s_nreloc;
				sct->head.s_nlnno   = scn.s_nlnno;
				sct->head.s_flags   = scn.s_flags;
				sct->section_nr     = i + 1;

				if(debugflag==1)
				{
					fprintf(stderr, "s_name    %*.*s\n",    S_SCNNMLEN, S_SCNNMLEN, scn.s_name);
					fprintf(stderr, "s_paddr   %08x\n",     scn.s_paddr);
					fprintf(stderr, "s_vaddr   %08x\n",     scn.s_vaddr);
					fprintf(stderr, "s_size    %08x\n",     scn.s_size);
					fprintf(stderr, "s_scnptr  %08x\n",     scn.s_scnptr);
					fprintf(stderr, "s_relptr  %08x\n",     scn.s_relptr);
					fprintf(stderr, "s_lnnoptr %08x\n",     scn.s_lnnoptr);
					fprintf(stderr, "s_nreloc  %08x\n",     scn.s_nreloc);
					fprintf(stderr, "s_nlnno       %04x\n", scn.s_nlnno);
					fprintf(stderr, "s_flags       %04x\n", scn.s_flags);
					fprintf(stderr, "\n");
				}

				break;
			}
		}

		if(sct->head.s_nreloc)
		{
			fseek(infile, sct->head.s_relptr, SEEK_SET);

			sct->relocate=calloc(sct->head.s_nreloc, sizeof(struct reloc));
			for(j=0;j<sct->head.s_nreloc;j++)
			{
				coff_reloc_fread(&sct->relocate[j], 1, infile);

				if(debugflag==1)
				{
					fprintf(stderr, "r_vaddr  %08x\n",     sct->relocate[j].r_vaddr);
					fprintf(stderr, "r_symndx %08x\n",     sct->relocate[j].r_symndx);
					fprintf(stderr, "r_offset %08x\n",     sct->relocate[j].r_offset);
					fprintf(stderr, "r_type       %04x\n", sct->relocate[j].r_type);
					fprintf(stderr, "r_stuff      %04x\n", sct->relocate[j].r_stuff);
					fprintf(stderr, "\n");
				}
			}
		}

		IFOPT(OPT_FORKDEBUG)
		{
			if(sct->head.s_nlnno)
			{
				sct->lnno = calloc(sct->head.s_nlnno, sizeof(struct lineno));
				fseek(infile, sct->head.s_lnnoptr, SEEK_SET);
				for(j=0;j<sct->head.s_nreloc;j++)
				{
					coff_lineno_fread(&sct->lnno[j], 1, infile);
				}
			}
		}
		
		if(options & OPT_ARGS && !strcmp(sct->head.s_name,".args"))
			sct->head.s_size=sbp_arglen;

		sct=sct->next;
	}
	
	fseek(infile, file_head.f_symptr, SEEK_SET);
	symtab=calloc(file_head.f_nsyms, sizeof(struct syment));
	for(i=0;i<file_head.f_nsyms;i++)
	{
		coff_syment_fread(&symtab[i], 1, infile);

		symtab[i].n_value/=sizeof(int32_t);

		if(debugflag==1)
		{
			if (symtab[i].n_zeroes)
			{
				fprintf(stderr, "n_name   %*.*s\n", N_SYMNMLEN, N_SYMNMLEN, symtab[i].n_name);
			}
			else
			{
				fprintf(stderr, "n_offset %08x\n",       symtab[i].n_offset);
			}
			fprintf(stderr, "n_value  %08x\n",       symtab[i].n_value);
			fprintf(stderr, "n_scnum      %04x\n",   symtab[i].n_scnum);
			fprintf(stderr, "n_type       %04x\n",   symtab[i].n_type);
			fprintf(stderr, "n_sclass       %02x\n", symtab[i].n_sclass);
			fprintf(stderr, "n_numaux       %02x\n", symtab[i].n_numaux);
			fprintf(stderr, "\n");
		}
	}
}

int set_startoffsets(void)
{
	int32_t i=0;
	int do_process;

	uint32_t start[10], org[10]={0,}, orgstart[10]={0,};
	int32_t mem_num;

	sct=sct_beg;
	start[0]=datastart;
	start[1]=0;

	while(sct!=NULL)
	{
		do_process = 0;
		fseek(infile, FILEHDR_SIZE + file_head.f_opthdr, SEEK_SET);
		for (i=0;i<file_head.f_nscns;++i)
		{
			struct scnhdr scn;

			coff_scnhdr_fread(&scn, 1, infile);

			if (!strcmp(scn.s_name,sct->head.s_name))
			{
				if(sct->head.s_flags==STYP_TEXT)
				{
					sct->head.s_paddr=sct->head.s_vaddr=textstart;
					put_size("textlen", sct->head.s_size);
				}
				else
				{
					do_process = 1;
				}

				break;
			}
		}

		if (do_process
		    || (options & OPT_ARGS
		        && !strcmp(sct->head.s_name, ".args")))
		{
			mem_num = sct->m_num;
			if(sct->head.s_paddr==UINT32_C(0xffffffff))
			{
				start[mem_num]=0;
				sct->head.s_paddr=0;
				sct->head.s_vaddr=0;
			}
			if(sct->head.s_paddr==UINT32_C(0))
			{
				org[mem_num]=0;
				sct->head.s_paddr=sct->head.s_vaddr=start[mem_num];
				start[mem_num]+=sct->head.s_size;
			}
			else
			{
				if(orgstart[mem_num]!=sct->head.s_vaddr)
				{
					orgstart[mem_num]=sct->head.s_vaddr;
					org[mem_num]=orgstart[mem_num];
				}
				sct->head.s_paddr=start[mem_num];
				sct->head.s_vaddr=org[mem_num];
				start[mem_num]+=sct->head.s_size;
				org[mem_num]+=sct->head.s_size;
			}

			if(debugflag==1)
			{
				fprintf(stderr,
					"s_name   %s\n"
					"s_paddr  %08x\n"
					"s_vaddr  %08x\n"
					"start    %08x\n"
					"orgstart %08x\n"
					"org      %08x\n\n",
					sct->head.s_name,
					sct->head.s_paddr,
					sct->head.s_vaddr,
					start[mem_num],
					orgstart[mem_num],
					org[mem_num]);
			}

		}

		if (sct->section_nr)
		{
			v_addr[sct->section_nr-1]=sct->head.s_vaddr;
		}

		sct=sct->next;
	}

/*
	for (i=0;i<file_head.f_nscns;++i)
		fprintf(stderr, "v_addr[%i]: 0x%08x\n", i, v_addr[i]);;
*/

	return(0);
}

inline void get_dbg(int32_t dbgptr, struct lineno **l_lnno)
{
	*l_lnno = &sct->lnno[dbgptr];
}

inline void get_reloc(int32_t rel_ptr, struct reloc **l_rel)
{
	*l_rel=&sct->relocate[rel_ptr];

	if(debugflag==1)
	{
		fprintf(stderr, "r_vaddr  %08x\n", (*l_rel)->r_vaddr);
	}
}

inline void get_symtab(int32_t s_index, struct syment **l_sym)
{
	*l_sym=&symtab[s_index];

	if(debugflag==1)
	{
		if ((*l_sym)->n_zeroes)
		{
			fprintf(stderr, "n_name   %*.*s\n", N_SYMNMLEN, N_SYMNMLEN, (*l_sym)->n_name);
		}
		else
		{
			fprintf(stderr, "n_offset %08x\n",       (*l_sym)->n_offset);
		}
	}
}

static void
printout(uint32_t data)
{
	fprintf(dumpfile, "0x%08"PRIx32"\n", data);
}

void do_relocation(void)
{
	int32_t relptr=0, dbgptr=0;
	int32_t sct_ptr;
	struct reloc *load_rel=NULL;
	struct syment *load_sym=NULL, *load_sym2=NULL;
	struct lineno *load_dbg=NULL;

	uint32_t rel_mask=UINT32_C(0xffffffff);
	uint32_t sign_mask=0x0;
	uint32_t rel_value;
	uint32_t sct_long;
	int32_t host_blk[BUFSIZE];
	int32_t i=0, j=0;
	char *src;
	struct section *sct_p;

	fseek(infile, FILEHDR_SIZE + file_head.f_opthdr + SCNHDR_SIZE*file_head.f_nscns, SEEK_SET);

	sct=sct_beg;

	while(sct!=NULL)
	{
		switch(sct->head.s_flags)
		{
		case STYP_TEXT:

			fputs(".CODE\n",dumpfile);
			if (options & OPT_COD2)
				fprintf(dumpfile,".SIZE 0x%08x\n", sct->head.s_size);
			fprintf(dumpfile,".ORIGIN 0x%08x\n",sct->head.s_paddr);
			break;
		case STYP_DATA:
			if(!strcmp(sct->head.s_name,".lsdata") ||
			   !strcmp(sct->head.s_name,".lpdata") ||
			   !strcmp(sct->head.s_name,".LDATA"))
			{
				fputs(".LDATA\n",dumpfile);
				if (options & OPT_COD2)
					fprintf(dumpfile,".SIZE 0x%08x\n", sct->head.s_size);
				fprintf(dumpfile,".ORIGIN 0x%08x\n",sct->head.s_paddr);
			}
			else
			{
				fputs(".GDATA\n",dumpfile);
				if (options & OPT_COD2)
					fprintf(dumpfile,".SIZE 0x%08x\n", sct->head.s_size);
				fprintf(dumpfile,".ORIGIN 0x%08x\n",sct->head.s_paddr);
			}
			break;
		case STYP_BSS:
			if(!strcmp(sct->head.s_name,".lddata"))
			{
				struct lditem *ld_ptr = ld_data;

				fputs(".GDATA\n",dumpfile);
				if (options & OPT_COD2)
					fprintf(dumpfile,".SIZE 0x%08x\n", sct->head.s_size);
				fprintf(dumpfile,".ORIGIN 0x%08x\n",sct->head.s_paddr);

				while(ld_ptr < (ld_data + LD_DATA_LEN) && ld_ptr->name != NULL)
					printout(ld_ptr++->value);

				sct_p=sct;
				while(sct_p!=NULL)
				{
					printout(sct_p->head.s_size);
					sct_p=sct_p->next;
				}
			}

		default:
			break;
		}
		
		relptr=0;
		if(sct->head.s_flags!=STYP_BSS && strcmp(sct->head.s_name, ".args"))
		{
			fseek(infile, sct->head.s_scnptr, SEEK_SET);
			if(sct->head.s_nreloc)
			{
				get_reloc(relptr, &load_rel);
				get_symtab(load_rel->r_symndx, &load_sym);
				//relptr++;
				IFOPT(OPT_FORKDEBUG)
					if(sct->head.s_flags==STYP_TEXT)
					{
						do
							get_dbg(dbgptr, &load_dbg);
						while(load_dbg->l_lnno!=0);
						
						get_symtab(load_dbg->l_symndx, &load_sym2);
						relptr++;
					}
			}
			sct_ptr=0;
			while(sct_ptr<sct->head.s_size)
			{
				bw_fread_be32(&sct_long, 1, infile);
				if(sct_ptr==load_rel->r_vaddr && relptr < sct->head.s_nreloc)
				{
				/*	rel_value=load_sym->n_value+load_rel->r_offset;  */
					rel_value=load_rel->r_offset;
					switch(load_rel->r_type)
					{
					case R_13ABS:
						rel_value+=v_addr[load_sym->n_scnum-1];
						rel_mask=REL_MASK13;
						sign_mask=SIGN_13;
						break;
					case R_13REL:
						rel_value-=sct_ptr;
						rel_mask=REL_MASK13<<1;
						sign_mask=0x0;
						break;
					case R_19ABS:
						rel_value+=v_addr[load_sym->n_scnum-1];
						rel_mask=REL_MASK19;
						sign_mask=SIGN_19;
						break;
					case R_19REL:
						rel_value-=sct_ptr;
						rel_mask=REL_MASK19<<1;
						sign_mask=0x0;
						break;
					case R_GETHI:
						rel_value+=v_addr[load_sym->n_scnum-1];
						rel_value>>=13;
						rel_mask=REL_MASK19;
						sign_mask=SIGN_19;
						break;
					case R_32DAT:
						rel_value+=v_addr[load_sym->n_scnum-1];
						rel_mask=REL_MASK32;
						sign_mask=0x0;
						break;
					default:
						break;
					}

					sct_long=((sct_long & rel_mask & ~sign_mask) |
							  (rel_value & ~rel_mask));

					relptr++;
					if (relptr < sct->head.s_nreloc)
					{
						get_reloc(relptr, &load_rel);
						get_symtab(load_rel->r_symndx, &load_sym);
					}
				}
				printout(sct_long);


				IFOPT(OPT_FORKDEBUG)
					if(sct->head.s_flags==STYP_TEXT &&
					   sct_ptr==load_sym2->n_value &&
					   dbgptr <= sct->head.s_nlnno && sct->head.s_nlnno>0 )
					{
						if(!strncmp(load_sym2->n_name, "IBP", 3)
						   && load_sym2->n_sclass == C_LINE)
							fprintf(dumpfile,"%s\n", load_sym2->n_name);
						
						do
							get_dbg(dbgptr, &load_dbg);
						while(load_dbg->l_lnno!=0);

						get_symtab(load_dbg->l_symndx, &load_sym2);
						dbgptr++;
					}
				
				sct_ptr++;
			}
		}
		if(options & OPT_ARGS && !strcmp(sct->head.s_name, ".args"))
		{
			if (sct->head.s_size)
			{
				fputs(".GDATA\n",dumpfile);
				if (options & OPT_COD2)
					fprintf(dumpfile,".SIZE 0x%08x\n", sct->head.s_size);
				fprintf(dumpfile,".ORIGIN 0x%08x\n",sct->head.s_paddr);

/*
				host_blk[0]=sbp_argc;
				host_blk[1]=sct->head.s_vaddr+3;
				host_blk[2]=sct->head.s_vaddr+sbp_arglen;
				host_blk[3]=host_blk[1]+sbp_argc;
				for(i=0;i<sbp_argc-1;i++)
					host_blk[i+4]=host_blk[i+3]+strlen(sbp_argv[i])+1;
				src=sbp_argv[0];
				i+=4;
				sbp_arglen-=i;
				while(src-sbp_argv[0]<sbp_arglen)
				{
					if(i<BUFSIZE)
						host_blk[i++]=*src++;
					else
					{
						for(i=0;i<BUFSIZE;i++)
							printout(host_blk[i]);
						i=0;
						host_blk[i++]=*src++;
					}
				}
				for(j=0;j<i;j++)
					printout(host_blk[j]);
				i=0;
*/
				{
					int clen;
					/* argc plus dummy */
					printout(0);
					printout(sbp_argc);
					/* base adress ov argv */
					printout(sct->head.s_vaddr + 3);

					/* argv[0] .. argv[argc - 1] */
					clen = sct->head.s_vaddr + 3 + sbp_argc;
					for (i = 0; i < sbp_argc; i++) {
						printout(clen);
						clen += strlen(sbp_argv[i]) + 1;
					}

					/* last NULL pointer */
					//printout(0);

					/* print strings */
					for (i = 0; i < sbp_argc; i++) {
						size_t j;
						for (j = 0; j < strlen(sbp_argv[i]); j++)
							printout(sbp_argv[i][j]);
						printout(0);
					}
				}
			}
		}

		sct=sct->next;
	}
}

#ifdef HOST
void communicate_host(int ret)
{
	int message_buffer[256];
	
	attach_queue(message_key, 1);
	message_buffer[0] = ret;
	if(ret<0)
	{
		send_msg(message_key, (char *)message_buffer, 1*sizeof(int));
		exit(0);
	}
	else
	{
		message_buffer[1] = get_section_size(0);
		put_size("textlen", message_buffer[1]);       /* Wozu ???? */
		message_buffer[2] = get_typesize(LD_PRIVATE) * get_typesize(LD_NUM)
			+ get_typesize(LD_SHARED)
			+ get_section_size(SBP_PRIVATE) * get_typesize(LD_NUM)
			+ get_section_size(SBP_SHARED)
		    + 515 * get_typesize(LD_NUM) + 1000;      /* istacksize ### */
		message_buffer[3] = get_typesize(LD_NUM);
	}
	send_msg(message_key, (char *)message_buffer, 4*sizeof(int));
	receive_msg(message_key, (char *)message_buffer);
	textstart = message_buffer[1];
	datastart = message_buffer[2];
	put_size("vpmask", message_buffer[3]);
}
#endif

int ld_data_calculate(void)
{
	int i=0;
	int gmem_size;
	
	gmem_size = get_typesize(LD_PRIVATE) * get_typesize(LD_NUM)
		    + get_typesize(LD_SHARED)
			+ get_section_size(SBP_PRIVATE) * get_typesize(LD_NUM)
			+ get_section_size(SBP_SHARED)
			+ 515 * get_typesize(LD_NUM);      /* istacksize ### */
	put_size("sheap", gmem_size);
	return 0;
}

void emit_memsize(void)
{
  int ProgramMemLen, GlobalMemLen;

  ProgramMemLen = textstart + get_section_size(0);

  GlobalMemLen = datastart + get_typesize(LD_PRIVATE) * get_typesize(LD_NUM)
	+ get_typesize(LD_SHARED)
	+ get_section_size(SBP_PRIVATE) * get_typesize(LD_NUM)
	+ get_section_size(SBP_SHARED)
	+ 515 * get_typesize(LD_NUM) + 1000;      /* istacksize ### */

  fputs(".CODE\n", dumpfile);
  if (options & OPT_COD2)
    fputs(".SIZE 0x0\n", dumpfile);
  fprintf(dumpfile,".ORIGIN 0x%08x\n", ProgramMemLen);
  fputs(".GDATA\n", dumpfile);
  if (options & OPT_COD2)
    fputs(".SIZE 0x0\n", dumpfile);
  fprintf(dumpfile,".ORIGIN 0x%08x\n", GlobalMemLen);
}


int main(int argc, char *argv[])
{
	int ret = 0;

	outfilename = NULL;
	memset(ld_data, 0, sizeof(ld_data));

	get_environment();

	configfile=fopen(".ldrc", "r");
	if(configfile==NULL)
	{
		configfile=fopen(loaderrc,"r");
	}
	if (!configfile && getenv("PRAMDIR"))
	{
		snprintf(loaderrc, sizeof(loaderrc), "%s/.ldrc", getenv("PRAMDIR"));
		configfile = fopen(loaderrc, "r");
	}
	if(configfile==NULL)
	{
		snprintf(loaderrc, sizeof(loaderrc), "%s/.ldrc", getenv("HOME"));
		configfile=fopen(loaderrc,"r");
	}
	if(configfile==NULL)
	{
		fprintf(stderr,"Warning couldn't open loaderrc %s\n", loaderrc);
		exit(1);
	}
	else
	    get_mem_values();

	eval_params(argc,argv);

	if(debugflag==1)
	{
		fprintf(stderr, "loaderrc: %s\n", loaderrc);
	}

	if(!(infile=fopen(infilename,"r")))
	{
	    fprintf(stderr, "Error, couldn't open inputfile: %s\n", infilename);
#ifndef HOST
		exit(-1);
#else
		ret = -1;
#endif
	}
	if (outfilename)
		dumpfile=fopen(outfilename,"w");
	else
		dumpfile=stdout;

	if(ret!=-1)
		scan_coff_file();
		
#ifdef HOST	
	IFOPT(OPT_HOST_COM)
		communicate_host(ret);
#endif

	emit_memsize();
	ld_data_calculate();
	
	set_startoffsets();
	do_relocation();
	if (dumpfile != stdout)
		fclose(dumpfile);
	//sync();
	
#ifdef HOST	
	IFOPT(OPT_HOST_COM)
		send_msg(message_key, argv[0], 1);
#endif
	
	return(0);
}
