#define _GNU_SOURCE
#include "config.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <assert.h>

#include "processor.h"
#include "glovars.h"
#include "simul.h"

#include "mem.h"

char *MemTraceFilename = NULL;
static FILE *memTraceFile;

void InitMemTrace(void)
{
	if (MemTraceFilename) {
		bool traceFilenameGenerated = false;

		if (!MemTraceFilename[0]
		    || !strcmp(MemTraceFilename, ":generate:")) {
			MemTraceFilename = GenerateFilename("trace");
			traceFilenameGenerated = true;
		}

		memTraceFile = fopen(MemTraceFilename, "w");

		if (traceFilenameGenerated) {
			free(MemTraceFilename);
			MemTraceFilename = ":generate:";
		}
	}
}

void FiniMemTrace(void)
{
	if (memTraceFile)
		fclose(memTraceFile);
	memTraceFile = NULL;
}

void StartMemTraceOutput(void)
{
	if (memTraceFile)
		fprintf(memTraceFile, "%08u:", PhysicalProcessor->CT);
}

void EndMemTraceOutput(void)
{
	if (memTraceFile)
		fprintf(memTraceFile, "\n");
}

void DoTraceMemRequest(struct vpregs *vp)
{
	struct memreq *mr = &vp->memreq;
	static const char *opNames[] = {
		[DO_OP_NONE] = "SET",
		[DO_OP_AND]  = "AND",
		[DO_OP_OR]   = "OR ",
		[DO_OP_ADD]  = "ADD",
		[DO_OP_MAX]  = "MAX"
	};
	static const char *modNames[] = {
		[MOD_LD] = "LD",
		[MOD_ST] = "ST",
		[MOD_MP] = "MP",
		[MOD_SY] = "SY"
	};


	if (!memTraceFile)
		return;

	if (mr->inUse) {
		if (mr->mod == MOD_MP || mr->mod == MOD_LD)
			fprintf(memTraceFile,
			        " "
#if defined(MEM_TRACE_WITH_PC) && MEM_TRACE_WITH_PC > 0
			        "0x%08x:"
#endif
			        "0x%08x:%02d:%c:0x%08x:0x%08x"
#if defined(MEM_TRACE_WITH_REGS) && MEM_TRACE_WITH_REGS > 0
			        ":%s:%s"
#endif
#if defined(MEM_TRACE_WITH_PC) && MEM_TRACE_WITH_PC > 0
			        , mr->PC
#endif
			        , mr->paddr, mr->regno, '1'
			        , mr->data,  mr->retval
#if defined(MEM_TRACE_WITH_REGS) && MEM_TRACE_WITH_REGS > 0
			        , modNames[mr->mod], opNames[mr->op]
#endif
			);
		else
			fprintf(memTraceFile,
			        " "
#if defined(MEM_TRACE_WITH_PC) && MEM_TRACE_WITH_PC > 0
			        "0x%08x:"
#endif
			        "0x%08x:%02d:%c:0x%08x:0x%08x"
#if defined(MEM_TRACE_WITH_REGS) && MEM_TRACE_WITH_REGS > 0
			        ":%s:%s"
#endif
#if defined(MEM_TRACE_WITH_PC) && MEM_TRACE_WITH_PC > 0
			        , mr->PC
#endif
			        , mr->paddr, mr->regno, '0'
			        , mr->data,  UINT32_C(0)
#if defined(MEM_TRACE_WITH_REGS) && MEM_TRACE_WITH_REGS > 0
			        , modNames[mr->mod], opNames[mr->op]
#endif
			);
	} else
		fprintf(memTraceFile, "%*s",
		       38
#if defined(MEM_TRACE_WITH_PC) && MEM_TRACE_WITH_PC > 0
		       + 11
#endif
#if defined(MEM_TRACE_WITH_REGS) && MEM_TRACE_WITH_REGS > 0
		       + 7
#endif
		      , ""
		);
}
