#define _GNU_SOURCE
#include "config.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <assert.h>

#include "processor.h"
#include "glovars.h"
#include "simul.h"

#include "mem.h"

static bool haveReq;

void IssueMemRequest(
	struct vpregs *vp,
	enum net_mod  mod,
	enum mem_op   op,
	UWORD         paddr,
	UWORD         vaddr,
	WORD          data,
	UWORD         SR,
	int           regno,
	int           force
)
{
	struct memreq *mr = &vp->memreq;

	mr->inUse = true;
	mr->mod   = mod;
	mr->op    = op;
	mr->paddr = paddr;
	mr->vaddr = vaddr;
	mr->data  = data;
	mr->SR    = SR;
	mr->regno = regno;
	mr->force = force;
	mr->PC    = vp->R[1];

	haveReq = true;
}

void GetMemRequestResult(struct vpregs *vp)
{
	struct memreq *mr = &vp->memreq;

	ResetStatMemRequest(vp);

	if (!mr->inUse)
		return;

	vp->dl_regno = mr->regno;
	vp->dl_val   = mr->retval;

	mr->inUse = false;

}

static void processMemRequest(struct vpregs *vp)
{
	WORD mem = 0;
	struct memreq *mr = &vp->memreq;

	if (!mr->inUse) {
		DoTraceMemRequest(vp);
		return;
	}

	mr->retval = mem = GlobalMem[mr->paddr];

	DoTraceMemRequest(vp);
	DoStatMemRequest(vp);

	mr->inUse = false;
	if (mr->mod == MOD_MP || mr->mod == MOD_LD) {
		mr->inUse = true;

		MarkReadStatMemRequest(vp);
	}

	if ((mr->mod != MOD_LD) && ((!(mr->SR & SHADOWFlag)) || mr->force)) {
		switch (mr->op) {
		case DO_OP_NONE:
			GlobalMem[mr->paddr] = mr->data;
			break;

		case DO_OP_AND:
			GlobalMem[mr->paddr] = mr->data & mem;
			break;

		case DO_OP_OR:
			GlobalMem[mr->paddr] = mr->data | mem;
			break;

		case DO_OP_ADD:
			GlobalMem[mr->paddr] = mr->data + mem;
			break;

		case DO_OP_MAX:
			GlobalMem[mr->paddr] = (mr->data > mem) ? mr->data: mem;
			break;
		}

		MarkWriteStatMemRequest(vp);
	}
}

void
MemInit(void)
{
	InitMemTrace();
	InitMemStats();
}

void
MemFini(void)
{
	FiniMemTrace();
	FiniMemStats();
}

void
SimulateMem(void)
{
	int i;

	if (haveReq) {
		StartMemTraceOutput();

		for (i = 0; i < PPnum*VPnum; i++) {
			processMemRequest(&VirtualProcessor[i]);
		}

		haveReq = false;
	}

	EndMemTraceOutput();

	OutputMemStats();
}
