/* -*-mb0-c-*- */
/* FILE:        coff.c
   DESCRIPTION: generating coff-coded files
   AUTHOR:      Stefan Franziskus
   DATE:        DEC 14, 1993
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <inttypes.h>

#include <bw/boio.h>

#include "define.h"
#include "coff.h"

struct section *sct;           /* Sectionliste */
static struct section *sct_beg;       /* Pointer auf Anfang der Sectionliste */
static struct sym_blk *symtab;
static struct sym_blk *symtab_beg;
static struct filehdr file_head;
static struct aouthdr aout_head;
extern FILE *outfile,*dumpfile;
static int32_t str_tab_len=sizeof(int32_t);

static inline void *
asmalloc(size_t size)
{
    void *m = malloc(size);

    if (!m) {
        fprintf(stderr, "no mem\n");
        exit(1);
    }

    memset(m, 0, size);

    return m;
}

void coff_init(PASSMODE runcnt)
{
	if (runcnt==FIRST_RUN || runcnt==SINGLE_RUN)
	{
		sct=NULL;
		sct_beg=NULL;
		file_head.f_opthdr=AOUTHDR_SIZE;
		symtab=asmalloc(sizeof(struct sym_blk));
		symtab_beg=symtab;
	}
	if (runcnt==SECOND_RUN)
	{
		sct=sct_beg;
		while(sct!=NULL)
		{
			sct->head.s_size=0;
			sct=sct->next;
		}
	}
}

static void add_section(const char *name, int32_t type)
{
	if (strlen(name) > S_SCNNMLEN)
		fprintf(stderr, "%s: warning: section name too long: %s\n", __func__, name);

	if(sct==NULL)
	{
		sct=asmalloc(sizeof(struct section));
		sct->section_nr=1;
	}
	else
	{
		sct->next=asmalloc(sizeof(struct section));
		sct->next->section_nr=sct->section_nr+1;
		sct=sct->next;
	}
	sct->next=NULL;

	file_head.f_nscns++;

	if(sct_beg==NULL)
		sct_beg=sct;
	/*
	while(*name && i < 8)
		sct->head.s_name[i++]=*name++;
	*/
	strncpy(sct->head.s_name, name, S_SCNNMLEN);
	sct->head.s_name[S_SCNNMLEN] = '\0';
	sct->symndx=file_head.f_nsyms;
	sct->head.s_size=0;
	sct->head.s_flags=type;
	sct->data=asmalloc(sizeof(struct section_blk));
	sct->data_start=sct->data;
	sct->head.s_nreloc=0;
	sct->relocate=asmalloc(sizeof(struct reloc_blk));
	sct->reloc_start=sct->relocate;
	sct->head.s_nlnno=0;
	sct->lineno=asmalloc(sizeof(struct lineno_blk));
	sct->lineno_start=sct->lineno;
}

int search_section(const char *name, int32_t type)
{
	sct=sct_beg;

	while(sct!=NULL && sct->next!=NULL && strcmp(sct->head.s_name,name))
		sct=sct->next;

	if(sct==NULL || (sct->next==NULL && strcmp(sct->head.s_name,name)))
		add_section(name,type);

	if(sct->head.s_flags!=type)
		return(1); /*error*/

	return(1);
}

int coff_code_out(int32_t code)
{
	if(sct==NULL)
	{
		errmsg(NO_SCT_ERR);
		return(0);
	}

	if ((sct->head.s_size-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
	{
		sct->data->next=asmalloc(sizeof(struct section_blk));
		sct->data=sct->data->next;
		sct->data->next=NULL;
	}

	/* we must convert section data, we are the generator */
	sct->data->entry[sct->head.s_size % COFF_BLK_SIZE] = bw_cpu_to_be32(code);

	return(1);
}

static int32_t search_syment(int16_t sct_num)
{
	int32_t block_ptr=0;
	struct sym_blk *symtab_last;
	int32_t return_value = 0;
  
	symtab_last=symtab;
	symtab=symtab_beg;
	while(block_ptr < file_head.f_nsyms && symtab->entry[block_ptr % COFF_BLK_SIZE].se.n_scnum!=sct_num)
	{
		block_ptr++;
		if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
			symtab=symtab->next;
	}
	if (symtab)
		return_value=symtab->entry[block_ptr % COFF_BLK_SIZE].se.n_value;
	else
		fprintf(stderr, "%s: warning: search for %"PRId16" failed\n", __func__, sct_num);
	symtab=symtab_last;
	return(return_value);
}

static int32_t get_symndx(int16_t sct_num)
{
	int32_t block_ptr=0;
	struct section *sct_ptr;
	struct sym_blk *symtab_last;

	sct_ptr=sct_beg;
	if(sct_num!=0)
	{
	  while(sct_ptr!=NULL && sct_ptr->section_nr!=sct_num)
	    sct_ptr=sct_ptr->next;
	  return(sct_ptr->symndx);
	}

	symtab_last=symtab;
	symtab=symtab_beg;
	while(block_ptr < file_head.f_nsyms && symtab->entry[block_ptr % COFF_BLK_SIZE].se.n_scnum!=sct_num)
	{
		block_ptr++;
		if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
			symtab=symtab->next;
	}

	symtab=symtab_last;
	return(block_ptr);
}

int create_relocent(int32_t vaddr, int16_t sct_num, int32_t symval, int32_t adr,  uint16_t type)
{
	uint32_t c_r_offset;
	int32_t symndx;

	symndx=get_symndx(sct_num);
	c_r_offset=symval-search_syment(sct_num);

	if (sct_num==0)
	{
		symndx=adr;
		c_r_offset=symval-adr;
	}

	if((sct->head.s_nreloc-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
	{
		sct->relocate->next=asmalloc(sizeof(struct reloc_blk));
		sct->relocate=sct->relocate->next;
		sct->relocate->next=NULL;
	}

	sct->relocate->entry[sct->head.s_nreloc % COFF_BLK_SIZE].r_vaddr=vaddr;
	sct->relocate->entry[sct->head.s_nreloc % COFF_BLK_SIZE].r_symndx=symndx;
	sct->relocate->entry[sct->head.s_nreloc % COFF_BLK_SIZE].r_offset=c_r_offset;
	sct->relocate->entry[sct->head.s_nreloc % COFF_BLK_SIZE].r_type=type;

	sct->head.s_nreloc++;

	return(0);
}

void create_lineno_ent(int32_t paddr, int32_t lnno)
{
	if((sct->head.s_nlnno-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
	{
		sct->lineno->next=asmalloc(sizeof(struct lineno_blk));
		sct->lineno=sct->lineno->next;
		sct->lineno->next=NULL;
	}

	sct->lineno->entry[sct->head.s_nlnno % COFF_BLK_SIZE].l_paddr=paddr;
	sct->lineno->entry[sct->head.s_nlnno % COFF_BLK_SIZE].l_lnno=lnno;
	sct->head.s_nlnno++;
}


int patch_syment(const char *name, int32_t value, int16_t scnum, uint8_t sclass)
{
	struct sym_blk *symtab_last;
	int32_t i=0;

	symtab_last=symtab;
	symtab=symtab_beg;

	while(i<file_head.f_nsyms && strcmp(symtab->entry[i% COFF_BLK_SIZE].name,name))
	{
		i++;
		if ((i-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
			symtab=symtab->next;
		
	}

	symtab->entry[i % COFF_BLK_SIZE].se.n_value=value;
	symtab->entry[i % COFF_BLK_SIZE].se.n_scnum=scnum;
	symtab->entry[i % COFF_BLK_SIZE].se.n_sclass=sclass;

	symtab=symtab_last;

	return(i);
}

int32_t create_syment(const char *name, int32_t value, int16_t scnum, uint8_t sclass)
{
	if ((file_head.f_nsyms-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
	{
		symtab->next=asmalloc(sizeof(struct sym_blk));
		symtab=symtab->next;
		symtab->next=NULL;
	}

	if(strlen(name)<=N_SYMNMLEN)
	{
		/*
		while(*name)
			symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_name[i++]=*name++;
		*/
		strncpy(symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_name, name, N_SYMNMLEN);
		symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_name[N_SYMNMLEN] = '\0';
		symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].name = symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_name;
	}
	else
	{
		symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_zeroes=0;
		symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_offset=str_tab_len;
		str_tab_len+=(strlen(name)+1);
		symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].name = strdup(name);
	}

	symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_value=value;
	symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_scnum=scnum;
	symtab->entry[file_head.f_nsyms % COFF_BLK_SIZE].se.n_sclass=sclass;

	file_head.f_nsyms++;

	return(file_head.f_nsyms-1);
}

static void calculate_pointers(void)
{
	int i;
	int32_t file_ptr;

	file_ptr=FILEHDR_SIZE+AOUTHDR_SIZE+file_head.f_nscns*SCNHDR_SIZE;
	aout_head.text_start=file_ptr;

	sct=sct_beg;

	while(sct!=NULL)
	{
		if(sct->head.s_flags!=STYP_BSS)
		{
			if(i==STYP_DATA)
				aout_head.data_start=file_ptr;

			sct->head.s_scnptr=file_ptr;
			file_ptr+=sct->head.s_size * sizeof(int32_t);
		}
		sct=sct->next;
	}

	sct=sct_beg;
	while(sct!=NULL)
	{
		sct->head.s_relptr=file_ptr;
		file_ptr+=sct->head.s_nreloc*RELOC_SIZE;

		sct=sct->next;
	}

	sct=sct_beg;
	while(sct!=NULL)
	{
		sct->head.s_lnnoptr=file_ptr;
		file_ptr+=sct->head.s_nlnno*LINENO_SIZE;

		sct=sct->next;
	}

	file_head.f_symptr=file_ptr;
}

int coff_file_out(void)
{
	int32_t block_ptr;
	int32_t block_size;

	sct=sct_beg;
	while(sct!=NULL)
	{
		switch(sct->head.s_flags)
		{
		case STYP_TEXT:
			aout_head.tsize+=(sct->head.s_size*sizeof(int32_t));
			break;
		case STYP_DATA:
			aout_head.dsize+=(sct->head.s_size*sizeof(int32_t));
			break;
		case STYP_BSS:
			aout_head.bsize+=(sct->head.s_size*sizeof(int32_t));
			break;
		default:
			break;
		}
		sct=sct->next;
	}

	calculate_pointers();

	file_head.f_magic=SBP_MAGIC;
	file_head.f_flags=F_LNNO | F_AR32W;

	coff_filehdr_fwrite(&file_head, 1, outfile);

	coff_aouthdr_fwrite(&aout_head, 1, outfile);

	sct=sct_beg;
	while(sct!=NULL)
	{
		struct scnhdr *scnhdr = &sct->head;
		scnhdr->s_size*=sizeof(int32_t);

		coff_scnhdr_fwrite(scnhdr, 1, outfile);

		sct=sct->next;
	}

	sct=sct_beg;
	while(sct!=NULL)
	{
		sct->data=sct->data_start;
		block_size=sct->head.s_size / sizeof(int32_t);

		if(sct->head.s_flags!=STYP_BSS)
		{
			while(block_size>0)
			{
				if(block_size>COFF_BLK_SIZE)
					block_ptr=COFF_BLK_SIZE;
				else
					block_ptr=block_size;

				fwrite(sct->data->entry, sizeof(int32_t), block_ptr, outfile);
				block_size-=COFF_BLK_SIZE;
				sct->data=sct->data->next;
			}
		}

		sct=sct->next;
	}

	sct=sct_beg;
	while(sct!=NULL)
	{
		sct->relocate=sct->reloc_start;
		for(block_ptr=0;block_ptr<sct->head.s_nreloc;block_ptr++)
		{
			struct reloc *reloc;

			if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
				sct->relocate=sct->relocate->next;

			reloc = &sct->relocate->entry[block_ptr % COFF_BLK_SIZE];

			coff_reloc_fwrite(reloc, 1, outfile);
		}
		sct=sct->next;
	}

	sct=sct_beg;
	while(sct!=NULL)
	{
		sct->lineno=sct->lineno_start;
		for(block_ptr=0;block_ptr<sct->head.s_nlnno;block_ptr++)
		{
			struct lineno *lineno;

			if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
				sct->lineno=sct->lineno->next;

			lineno = &sct->lineno->entry[block_ptr % COFF_BLK_SIZE];

			coff_lineno_fwrite(lineno, 1, outfile);
		}
		sct=sct->next;
	}

	symtab=symtab_beg;
	for(block_ptr=0;block_ptr<file_head.f_nsyms;block_ptr++)
	{
		struct sym *syment;

		if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
			symtab=symtab->next;

		syment = &symtab->entry[block_ptr % COFF_BLK_SIZE];

		coff_syment_fwrite(&syment->se, 1, outfile);
	}

	bw_fwrite_be32(&str_tab_len, 1, outfile);

	symtab=symtab_beg;
	for(block_ptr=0;block_ptr<file_head.f_nsyms;block_ptr++)
	{
		struct sym *syment;

		if ((block_ptr-1) % COFF_BLK_SIZE+1 >= COFF_BLK_SIZE)
			symtab=symtab->next;

		syment = &symtab->entry[block_ptr % COFF_BLK_SIZE];

		if(syment->se.n_zeroes==0)
		{
			fwrite(syment->name, sizeof(char), strlen(syment->name) + 1, outfile);
		}
	}

	fflush(outfile);

	return(0);
}
